//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Util/UtilXML.cpp
//! @brief     Implements
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Util/UtilXML.h"
#include "Base/Util/Assert.h"
#include <QColor>
#include <QUuid>

namespace {

void assertCurrentTag(QXmlStreamReader* reader, const QString& expectedTag)
{

    if (reader->name() != expectedTag)
        throw std::runtime_error("Found tag '" + reader->name().toString().toStdString()
                                 + "' instead of expected expected tag '"
                                 + expectedTag.toStdString() + "'");
}

void assertCurrentToken(QXmlStreamReader* reader, QXmlStreamReader::TokenType token)
{
    if (reader->tokenType() != token)
        throw std::runtime_error("Found unexpected token type");
}

} // namespace


void XML::gotoEndElementOfTag(QXmlStreamReader* reader, const QString& tag)
{
    ASSERT(reader);
    if (reader->name() != tag) {
        if (!reader->isEndElement())
            reader->skipCurrentElement();
        reader->skipCurrentElement();
    }
    assertCurrentTag(reader, tag);
    if (!reader->isEndElement())
        reader->skipCurrentElement();

    assertCurrentToken(reader, QXmlStreamReader::EndElement);
    assertCurrentTag(reader, tag);
}

// ------------ write atribute ------------

void XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName, bool b)
{
    writer->writeAttribute(attributeName, b ? "1" : "0");
}

void XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName, double d)
{
    writer->writeAttribute(attributeName, d == 0.0 ? "0" : QString::number(d, 'e', 12));
}

void XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName, QString s)
{
    writer->writeAttribute(attributeName, s);
}

void XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName, QColor c)
{
    writer->writeAttribute(attributeName, c.isValid() ? c.name(QColor::HexArgb) : "");
}

void XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeBaseName, const R3& vec)
{
    writeAttribute(writer, attributeBaseName + "X", vec.x());
    writeAttribute(writer, attributeBaseName + "Y", vec.y());
    writeAttribute(writer, attributeBaseName + "Z", vec.z());
}

void XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeBaseName,
                         const complex_t& c)
{
    writeAttribute(writer, attributeBaseName + "Re", c.real());
    writeAttribute(writer, attributeBaseName + "Im", c.imag());
}


// ------------ read atribute ------------

bool XML::readBool(QXmlStreamReader* reader, const QString& attributeName)
{
    return reader->attributes().value(attributeName).toUInt() > 0;
}

int XML::readInt(QXmlStreamReader* reader, const QString& attributeName)
{
    return reader->attributes().value(attributeName).toInt();
}

unsigned XML::readUInt(QXmlStreamReader* reader, const QString& attributeName)
{
    return reader->attributes().value(attributeName).toUInt();
}

size_t XML::readSize(QXmlStreamReader* reader, const QString& attributeName)
{
    return reader->attributes().value(attributeName).toULong();
}

double XML::readDouble(QXmlStreamReader* reader, const QString& attributeName)
{
    return reader->attributes().value(attributeName).toDouble();
}

QString XML::readString(QXmlStreamReader* reader, const QString& attributeName)
{
    return reader->attributes().value(attributeName).toString();
}

QColor XML::readColor(QXmlStreamReader* reader, const QString& attributeName)
{
    const QString parameter_value = reader->attributes().value(attributeName).toString();
    return {parameter_value};
}

R3 XML::readR3(QXmlStreamReader* reader, const QString& attributeBaseName)
{
    double x = readDouble(reader, attributeBaseName + "X");
    double y = readDouble(reader, attributeBaseName + "Y");
    double z = readDouble(reader, attributeBaseName + "Z");
    return {x, y, z};
}

complex_t XML::readC(QXmlStreamReader* reader, const QString& attributeBaseName)
{
    double r = readDouble(reader, attributeBaseName + "Re");
    double i = readDouble(reader, attributeBaseName + "Im");
    return {r, i};
}
