//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Plotter/PlotStatusLabel.cpp
//! @brief     Implements class PlotStatusLabel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Plotter/PlotStatusLabel.h"
#include "GUI/View/Base/Fontsize.h"
#include "GUI/View/Plotter/ScientificPlot.h"
#include <QColor>
#include <QFont>
#include <QPainter>

namespace {

int default_text_size()
{
    return GUI::Style::fontSizeRegular();
}
int default_label_height()
{
    return GUI::Style::SizeOfLetterM().height() * 1.75;
}

} // namespace


PlotStatusLabel::PlotStatusLabel(ScientificPlot* plot)
    : m_font("Monospace", default_text_size(), QFont::Normal, false)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    setFixedHeight(default_label_height());

    if (plot)
        addPlot(plot);
}

void PlotStatusLabel::setText(const QString& text)
{
    m_text = text;
    update();
}

void PlotStatusLabel::paintEvent(QPaintEvent* event)
{
    QWidget::paintEvent(event);

    QPainter painter(this);
    painter.setBrush(QColor(Qt::black));
    painter.setPen(QColor(Qt::black));
    painter.setFont(m_font);

    QRect textRect(0, 0, geometry().width(), geometry().height());
    painter.fillRect(textRect, QColor(Qt::white));
    painter.drawText(textRect, m_alignment, m_text);
}

//! Disconnects all color maps from the label.
void PlotStatusLabel::reset()
{
    for (auto* colorMap : m_plots)
        setPlotLabelEnabled(colorMap, false);

    m_plots.clear();
}

void PlotStatusLabel::addPlot(ScientificPlot* plot)
{
    if (m_plots.contains(plot))
        return;

    m_plots.push_back(plot);
    setPlotLabelEnabled(plot, true);
}

void PlotStatusLabel::setPlot(ScientificPlot* plot)
{
    reset();
    addPlot(plot);
}

//! Enables/disables label. If disabled, all colorMaps are disconnected and label is hiding.
void PlotStatusLabel::setLabelEnabled(bool flag)
{
    for (auto* colorMap : m_plots)
        setPlotLabelEnabled(colorMap, flag);

    setEnabled(flag);
}

//! Enables/disables showing of label for given plot.
void PlotStatusLabel::setPlotLabelEnabled(ScientificPlot* plot, bool flag)
{
    plot->setMouseTrackingEnabled(flag);
    setConnected(plot, flag);
}

//! Connects with colorMap's status string signal.
void PlotStatusLabel::setConnected(ScientificPlot* plot, bool flag) const
{
    if (flag) {
        connect(plot, &ScientificPlot::statusString, this, &PlotStatusLabel::setText,
                Qt::UniqueConnection);
        connect(plot, &ScientificPlot::destroyed, this, &PlotStatusLabel::onPlotDestroyed,
                Qt::UniqueConnection);
    } else {
        disconnect(plot, &ScientificPlot::statusString, this, &PlotStatusLabel::setText);
    }
}

void PlotStatusLabel::onPlotDestroyed(QObject* obj)
{
    auto it = std::remove_if(m_plots.begin(), m_plots.end(),
                             [obj](ScientificPlot* cm) { return cm == obj; });
    m_plots.erase(it, m_plots.end());
}
